#include "LatexMathExpression.h"

#include "FileLogger.h"

LatexMathExpression::LatexMathExpression(const std::string& expression)
	: mRawBody(expression),
		mType(Type::LINEAR)
{
	MLOG_INFO("Adding expression with content: " << expression);
	parse();
}

LatexMathExpression::LatexMathExpression(Type type)
	: mType(type)
{

}

LatexMathExpression::~LatexMathExpression()
{

}

void LatexMathExpression::onTagBodyChar(char c)
{
	if (c == '}')
	{
		onRBrace();
	}
	else
	{
		mWorkingString += c;
	}
}

void LatexMathExpression::onFreeChar(char c)
{
	if (c == '\\')
	{
		mLineState = LineState::IN_TAG_NAME;
	}
	else if (c == '{')
	{
		onLBrace();
	}
	else if (c == '}')
	{
		onRBrace();
	}
	else if (c == '^')
	{
		onSuperscript();
	}
	else if (c == '_')
	{
		onSubscript();
	}
	else if (std::isblank(c))
	{
		onSpace();
	}
	else
	{
		mWorkingString += c;
	}
}

void LatexMathExpression::parse()
{
	mLineState = LineState::NONE;
	mWorkingString.clear();

	for (auto c : mRawBody)
	{
		if (mLineState == LineState::IN_TAG_BODY)
		{
			onTagBodyChar(c);
		}
		else
		{
			onFreeChar(c);
		}
	}
	onSpace();
	onCloseExpression();
}

void LatexMathExpression::onSpace()
{
	if (mLineState == LineState::IN_TAG_NAME)
	{
		onFinishedTagName();
	}
	else
	{
		onLiteral();
	}
	mWorkingString.clear();
}

void LatexMathExpression::onLBrace()
{
	if (mLineState == LineState::IN_TAG_NAME)
	{
		onFinishedTagName();
	}
	else if (mLineState == LineState::AWAITING_LBRACE)
	{
		mLineState = LineState::IN_TAG_BODY;
	}
	else if (mLineState == LineState::IN_TAG_BODY)
	{
		mOpenBraceCount++;
	}
	else
	{
		mWorkingString += "{";
	}
}

void LatexMathExpression::onRBrace()
{
	if (mLineState == LineState::IN_TAG_BODY)
	{
		if (mOpenBraceCount == 0)
		{
			onTagBody();
		}
		else
		{
			mOpenBraceCount--;
			mWorkingString += "}";
		}
	}
	else
	{
		mWorkingString += "}";
	}
}

void LatexMathExpression::setRawContent(const std::string& content)
{
	auto expression = std::make_unique<LatexMathExpression>(content);
	mExpressions.push_back(std::move(expression));
}

void LatexMathExpression::onTagBody()
{
	onCloseExpression();
}

void LatexMathExpression::onSuperscript()
{

}

void LatexMathExpression::onSubscript()
{

}

void LatexMathExpression::onLiteral()
{
	LatexMathSymbol symbol{ LatexMathSymbol::Type::RAW, mWorkingString, mWorkingString };
	mWorkingSymbols.push_back(symbol);
}

void LatexMathExpression::onFinishedTagName()
{
	if (auto lookup = LatexSymbolLookup::getKnownTag(mWorkingString); lookup)
	{
		if (mLineState == LineState::AWAITING_LBRACE || mLineState == LineState::IN_TAG_NAME)
		{
			onCloseExpression();

			if (*lookup == LatexMathSymbol::Tag::FRAC)
			{
				mWorkingType = Type::FRACTION;
				auto expression = std::make_unique<LatexMathExpression>(mWorkingType);
				mWorkingExpression = expression.get();
				mExpressions.push_back(std::move(expression));
				mWorkingString.clear();
			}

			mLineState = LineState::IN_TAG_BODY;		
		}
	}
	else if (auto lookup = LatexSymbolLookup::getSymbolUtf8(mWorkingString); lookup)
	{
		LatexMathSymbol symbol{ LatexMathSymbol::Type::TAG, mWorkingString, *lookup };
		mWorkingSymbols.push_back(symbol);
		mLineState = LineState::NONE;
	}
}

void LatexMathExpression::onCloseExpression()
{
	if (mWorkingType == Type::LEAF)
	{
		if (!mWorkingSymbols.empty())
		{
			auto expression = std::make_unique<LatexMathExpression>(mWorkingType);
			expression->setContent(mWorkingSymbols);
			mExpressions.push_back(std::move(expression));
		}
	}
	else if (mWorkingType == Type::FRACTION)
	{
		mWorkingExpression->setRawContent(mWorkingString);

		if (mWorkingExpression->getExpressions().size() == 2)
		{
			mWorkingType = Type::LEAF;
			mLineState = LineState::NONE;
		}
		else
		{
			mLineState = LineState::AWAITING_LBRACE;
			mOpenBraceCount = 0;
		}
	}

	mWorkingString.clear();
	mWorkingSymbols.clear();
}

void LatexMathExpression::setContent(std::vector<LatexMathSymbol>& symbols)
{
	mSymbols = symbols;
}

const std::vector<LatexMathSymbol>& LatexMathExpression::getSymbols() const
{
	return mSymbols;
}

const LatexMathExpression::Type LatexMathExpression::getType() const
{
	return mType;
}

const std::vector<LatexMathExpressionPtr>& LatexMathExpression::getExpressions() const
{
	return mExpressions;
}